/*
    slforum version 1.x - Simple chat room usable via telnet.
    Copyright (C) 1999,2000,2008 Jeremy D. Monin, Dan Chin, Matt Lochansky
	(forum@shadowlands.org).
    Based on jdkchat version 1.3, J.D.Koftinoff Software, Ltd. (GPL'd)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __SLFORUMUTIL_H
#define __SLFORUMUTIL_H

/* from slforumutil.c */
#ifdef __cplusplus
extern "C" {
#endif

extern char telnet_init[];
extern char telnet_local_on[];
extern char telnet_local_off[];
extern char telnet_double_iac[];

    /* write first sz bytes of a buffer to socket fd */
    int lpwrite( int fd, const char *buf, int sz );

    /* write a short file (up to 2047 bytes) to socket fd;
     * on error, returns -1 and sets perror.  if fd = -1,
     * returns 0.
     */
    int lpcat( int fd, const char *fname );

    /* write, printf-style, a formatted buffer to socket fd;
     * up to 255 bytes are printed, # bytes printed is returned.
     */
    int lpprintf (int fd, const char *fmt, ...);  

    /** returns NULL if file does not exist,
     *  or if it's a symlink, directory, or block device.
     *  'desc' should be what it is, such as "log file" or "banner".
     */
    FILE *slf_safe_fopen (const char *desc,
			  const char *fname, const char *mode);

    /** returns 1 on error. (couldn't allocate memory, etc.) 
     *  On success, fills in the following fields of ddif:
     *  tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec.
     *  The contents of each of these fields is the number of
     *  units elapsed between fromtime and tomtime (tm_year is
     *  the number of years, tm_mon is the number of additional
     *  months, etc.)
     */
    int datediff (time_t fromtime, time_t totime, struct tm *ddif);

    /** Write time to a buffer.
     *  "incldate" 0=time; 1=time,date; 2=time,date,year.
     */
    void _writetime (char *buf, size_t bufsize, time_t whicht, int incldate);

    /**  reset slf_localtime's vars.
     */
    void slf_localtime_init ();

    /** Localtime which caches last result, for performance and to
     *  avoid reentrancy.  This fixes a halt/crash seen in gdb for glibc
     *  or linux kernel, 20081206.  Call slf_localtime_init before first use.
     *  If timep is NULL, use current time.
     */
    struct tm *slf_localtime (const time_t *timep);

#ifdef __cplusplus
}	/* end extern "C" */
#endif

#endif	/* ifndef __SLFORUMUTIL_H */
